DEVICE_TYPE = "STELLA-AQ"
SOFTWARE_VERSION_NUMBER = "1.7.0"

# STELLA Air Quality instrument
# Science and Technology Education for Land/ Life Assessment
# NASA open source software license
# Paul Mirel 2023/2024

print( "DEVICE_TYPE = {}".format( DEVICE_TYPE ))
print( "SOFTWARE_VERSION_NUMBER = {}".format( SOFTWARE_VERSION_NUMBER ))

# this software will work with any combination of these three sensors:
# PMSA003I air quality (particulates) sensor
# SCD-40 carbon dioxide sensor
# BME280 ambient temperature, pressure, humidity sensor

# load libraries
import gc
gc.collect()
start_mem_free_kB = gc.mem_free()/1000
print("start memory free {} kB".format( start_mem_free_kB ))

import os
import microcontroller
import board
import digitalio
import time
import rtc
import neopixel
import storage
import busio
from adafruit_pcf8523 import pcf8523
from adafruit_bme280 import basic as adafruit_bme280
import adafruit_scd4x
from adafruit_pm25.i2c import PM25_I2C
import displayio
import terminalio
from adafruit_display_text import label
import vectorio # for shapes
import sdcardio
import adafruit_max1704x

try:
    import adafruit_ssd1327
    alt_display = False
    display_type = "128x128"
except:
    import adafruit_displayio_ssd1306
    alt_display = True
    display_type = "128x64"
print( "Display type = {}".format( display_type ))

print( "Default processor type is: Thing Plus RP2040" )
sd_type = 1

DAYS = { 0:"Sunday", 1:"Monday", 2:"Tuesday", 3:"Wednesday", 4:"Thursday", 5:"Friday", 6:"Saturday" }
DATA_FILE = "/sd/data.csv"
LOW_BATTERY_VOLTAGE = 3.4
RED =   ( 0, 255, 0 )
GREEN = ( 255, 0, 0 )
BLUE =  ( 0, 0, 255 )
ORANGE = ( 24, 64, 0 )
OFF =   ( 0,0,0 )

def main():
    null_time = time.struct_time(( 2000,  01,   01,   00,  00,  00,   0,   -1,    -1 ))
    batch_number = 0
    displayio.release_displays()

    UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
    print("unique identifier (UID) : {0}".format( UID ))

    # initialize on board LED and on board neopixel
    onboard_LED = digitalio.DigitalInOut( board.LED )
    onboard_LED.direction = digitalio.Direction.OUTPUT
    # flash to say hello
    for i in range( 3 ):
        onboard_LED.value = True
        time.sleep( 0.1 )
        onboard_LED.value = False
        time.sleep( 0.1 )
    onboard_LED.value = True

    on_board_neopixel_pin = board.NEOPIXEL
    num_pixels = 1
    ORDER = neopixel.RGB
    pixels = neopixel.NeoPixel( on_board_neopixel_pin, num_pixels, brightness=0.2, auto_write=True, pixel_order=ORDER )

    # initialize SD card storage
    try:
        if sd_type == 0:
            sdcard = sdioio.SDCard(
                clock=board.SDIO_CLOCK,
                command=board.SDIO_COMMAND,
                data=board.SDIO_DATA,
                frequency=25000000)
            print( "sdioio success" )
        if sd_type == 1:
            cs = board.SD_CS
            sd_card_spi = busio.SPI(board.SD_SCK, MOSI=board.SD_MOSI, MISO=board.SD_MISO)
            sdcard = sdcardio.SDCard(sd_card_spi, cs)
            print( "sdcardio success" )
        print( "initalized sd card" )
        vfs = storage.VfsFat(sdcard)
        print( "established vfs" )
    except( OSError, ValueError ) as err:
        print( "No SD card found, or card is full: {}".format(err) )
        vfs = False
        pixels.fill( RED )

    header = ( "device_type, software_version, UID, batch_number, timestamp_iso8601, decimal_hour, "
        "carbon_dioxide_sensor_stale_data_boolean, carbon_dioxide_ppm, carbon_dioxide_uncertainty_ppm, "
        "carbon_dioxide_sensor_air_temperature_C, carbon_dioxide_sensor_air_temperature_uncertainty_C, "
        "carbon_dioxide_sensor_relative_humidity_percent, carbon_dioxide_sensor_relative_humidity_uncertainty_percent, "
        "air_temperature_C, air_temperature_uncertainty_C, relative_humidity_percent, relative_humidity_uncertainty_percent, "
        "barometric_pressure_hPa, barometric_pressure_uncertainty_hPa, altitude_uncalibrated_m, altitude_uncalibrated_uncertainty_m, "
        "air_quality_index_instantaneous_for_particulates, "
        "particle_1_um_concentration_ug_per_m^3, particle_2.5_um_concentration_ug_per_m^3, "
        "particle_10_um_concentration_ug_per_m^3, particle_concentration_uncertainty_ug_per_m^3, "
        "particle_0.3_um_count_per_100mL, particle_0.5_um_count_per_100mL, particle_1_um_count_per_100mL, "
        "particle_2.5_um_count_per_100mL, particle_10_um_count_per_100mL, particle_100_um_count_per_100mL, "
        "battery_voltage, battery_percentage"
        "\n" )
    CO2_uncertainty = "50ppm+5%_of_reading"
    CO2_sensor_air_temperature_uncertainty = 6
    CO2_sensor_relative_humidity_uncertainty = 1
    WX_sensor_air_temperature_uncertainty = 1
    WX_sensor_relative_humidity_uncertainty = 3
    WX_sensor_barometric_pressure_uncertainty = 1
    WX_sensor_altitude_uncertainty = 500
    particle_concentration_uncertainty_ug = 10

    desired_sample_interval_h = 1.0
    number_of_datapoints_to_average = 10
    if vfs:
        storage.mount(vfs, "/sd")
        initialize_data_file( header )
        desired_sample_interval_h, number_of_datapoints_to_average = check_STELLA_config()
        print( "desired_sample_interval_h = {}, number_of_datapoints_to_average = {}".format( desired_sample_interval_h, number_of_datapoints_to_average ))

    # initialize devices
    try:
        i2c_bus = busio.I2C(board.SCL, board.SDA, frequency=100000) # need slower (100kHz) i2c bus speed for PM25 sensor
    except ValueError:
        i2c_bus = False
        print( "i2c bus fail -- press reset button, or power off to restart" )
        pixels.fill( ORANGE )

    if i2c_bus:
        try:
            vbat_voltage_pin = AnalogIn(board.VOLTAGE_MONITOR)
        except (NameError, AttributeError):
            print( "analog voltage monitor not available" )
            vbat_voltage_pin = False

        try:
            monitor = adafruit_max1704x.MAX17048(i2c_bus)
        except:
            print( "fuel gauge battery monitor not available" )
            monitor = False

        if vbat_voltage_pin:
            print( "getting analog voltage reading" )
            battery_voltage = get_voltage(vbat_voltage_pin)
            battery_percentage = 0

        if monitor:
            print( "getting fuel gauge battery health" )
            battery_voltage = monitor.cell_voltage          #wake up the fuel gauge chip with a couple of throw away commands
            battery_percentage = monitor.cell_percent
            time.sleep(0.1)
            battery_voltage = monitor.cell_voltage
            battery_percentage = monitor.cell_percent
        # initialize clock
        try:
            real_time_clock = pcf8523.PCF8523(i2c_bus)
            clock_battery_low = real_time_clock.battery_low
            if clock_battery_low:
                print( "clock battery is low. replace clock battery" )
            else:
                print( "clock battery is OK" )
            timenow = real_time_clock.datetime
            if timenow.tm_wday not in range ( 0, 7 ):
                                    #(( year, mon, date, hour, min, sec, wday, yday, isdst ))
                t = time.struct_time(( 2023,  01,   01,   00,  00,  00,   0,   -1,    -1 ))
                real_time_clock.datetime = t
            timenow = real_time_clock.datetime
        except ValueError:
            print( "real time clock not available")
            timenow = null_time
            real_time_clock = False
        DAYS = ("Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday")
        print( "The date is %s %d-%d-%d" % ( DAYS[ timenow.tm_wday ], timenow.tm_year, timenow.tm_mon, timenow.tm_mday ))
        print( "The time is %d:%02d:%02d" % ( timenow.tm_hour, timenow.tm_min,timenow.tm_sec ))
        if real_time_clock:
            system_clock = rtc.RTC()
            system_clock.datetime = real_time_clock.datetime

        batch_number = update_batch( timenow )
        # initialize WX
        try:
            WX_sensor = adafruit_bme280.Adafruit_BME280_I2C( i2c_bus )
        except ValueError:
            WX_sensor = False

        # initialize CO2
        try:
            CO2_sensor = adafruit_scd4x.SCD4X( i2c_bus )
            print( "CO2_sensor serial number:", [hex(i) for i in CO2_sensor.serial_number] )
            # DISABLE SENSOR HERE:
            CO2_sensor.start_periodic_measurement()
            CO2_reading = ( 0, 0, 0, 0 )
            last_CO2_reading = CO2_reading
        except ValueError:
            CO2_sensor = False

        # initialize PM25
        try:
            reset_pin = None
            pm25_sensor = PM25_I2C( i2c_bus, reset_pin )
            pm25_reading = (0,0,0,0,0,0,0,0,0,0,0,0,0)
            last_pm25_reading = pm25_reading
        except (ValueError, RuntimeError):
            pm25_sensor = False

        # check for each device present. if not present, do not read, do not record that datum.
        # header should be for all devices, so the header works for a file even if a sensor has been added or subtracted

        # initialize display
        try:
            display_bus = displayio.I2CDisplay( i2c_bus, device_address=0x3D )
            if display_type == "128x128":
                WIDTH = 128
                HEIGHT = 128
                FONTSCALE = 1
                display = adafruit_ssd1327.SSD1327( display_bus, width=WIDTH, height=HEIGHT, rotation=0 )
            elif display_type == "128x64":
                WIDTH = 128
                HEIGHT = 64  # Change to 64 if needed
                display = adafruit_displayio_ssd1306.SSD1306(display_bus, width=WIDTH, height=HEIGHT)

            display_group = displayio.Group()
            display.show( display_group )
        except ValueError:
            display = False

    if display:
        WIDTH = 128
        HEIGHT = 128
        if alt_display:
            HEIGHT = 64

        BORDER = 2
        border_color = 0xFFFFFF
        border = displayio.Palette(1)
        border[0] = border_color
        outer_rectangle = vectorio.Rectangle(pixel_shader=border, width=WIDTH, height=HEIGHT, x=0, y=0)
        display_group.append( outer_rectangle )
        front_color = 0x000000
        front = displayio.Palette(1)
        front[0] = front_color
        inner_rectangle = vectorio.Rectangle(pixel_shader=front, width=128-2*BORDER, height=HEIGHT-2*BORDER, x=0+BORDER, y=0+BORDER)
        display_group.append( inner_rectangle )


        if not alt_display:
            name_group = displayio.Group( scale=3, x=10, y=25 )
            name_text = "STELLA"
            name_text_area = label.Label( terminalio.FONT, text=name_text, color=0xFFFFFF )
            name_group.append( name_text_area )
            display_group.append( name_group )

            subname_group = displayio.Group( scale=3, x=30, y=60 ) #x=45, y=60 )
            subname_text = "-AQ-"
            subname_text_area = label.Label( terminalio.FONT, text=subname_text, color=0xFFFFFF )
            subname_group.append( subname_text_area )
            display_group.append( subname_group )

            UID_group = displayio.Group( scale=2, x=10, y=90 )
            UID_text = "UID: " + str(UID)
            UID_text_area = label.Label( terminalio.FONT, text=UID_text, color=0xFFFFFF )
            UID_group.append( UID_text_area )
            display_group.append( UID_group )

            version_group = displayio.Group( scale=1, x=20, y=113 )
            version_text = "Version: " + SOFTWARE_VERSION_NUMBER
            version_text_area = label.Label( terminalio.FONT, text=version_text, color=0xFFFFFF )
            version_group.append( version_text_area )
            display_group.append( version_group )
            print( "display welcome screen" )



            colors = [RED, GREEN, BLUE]
            for i in range (len(colors)):
                pixels.fill(colors[i])
                time.sleep( 0.75 )

            for i in range( 4 ):
                display_group.pop()

            del name_group
            del name_text
            del name_text_area

            del subname_group
            del subname_text
            del subname_text_area

            del UID_group
            del UID_text
            del UID_text_area

            del version_group
            del version_text
            del version_text_area

            gc.collect()
            timenow = real_time_clock.datetime

            date_group = displayio.Group( scale=2, x=5, y=15 )
            date_text = ("{}-{:02d}-{:02d}".format( timenow.tm_year, timenow.tm_mon, timenow.tm_mday ))
            date_text_area = label.Label( terminalio.FONT, text=date_text, color=0xFFFFFF )
            date_group.append( date_text_area )
            display_group.append( date_group )

            time_group = displayio.Group( scale=2, x=11, y=40 )
            time_text = ("{:02d}:{:02d}:{:02d}Z".format( timenow.tm_hour, timenow.tm_min, timenow.tm_sec ))
            time_text_area = label.Label( terminalio.FONT, text=time_text, color=0xFFFFFF )
            time_group.append( time_text_area )
            display_group.append( time_group )

            batch_group = displayio.Group( scale=2, x=15, y=64 )
            batch_text = "batch {}".format(batch_number)
            batch_text_area = label.Label( terminalio.FONT, text=batch_text, color=0xFFFFFF )
            batch_group.append( batch_text_area )
            display_group.append( batch_group )

            battery_voltage, battery_percentage = check_battery( vbat_voltage_pin, monitor )
            if battery_voltage < LOW_BATTERY_VOLTAGE:
                main_battery_status = "LOW"
            elif battery_voltage > 4.0:
                main_battery_status = "FULL"
            else:
                main_battery_status = " OK"
            if battery_percentage:
                main_battery_status = ( "{}%".format( int( battery_percentage )))


            main_battery_group = displayio.Group( scale=2, x=10, y=89 )
            main_battery_text = ("Vbat:{}".format(main_battery_status))
            main_battery_text_area = label.Label( terminalio.FONT, text=main_battery_text, color=0xFFFFFF )
            main_battery_group.append( main_battery_text_area )
            display_group.append( main_battery_group )
            if clock_battery_low:
                clock_battery_status = "LOW"
            else:
                clock_battery_status = "OK"
            clock_battery_group = displayio.Group( scale=1, x=10, y=113 )
            clock_battery_text = ("clock battery: {}".format( clock_battery_status ))
            clock_battery_text_area = label.Label( terminalio.FONT, text=clock_battery_text, color=0xFFFFFF )
            clock_battery_group.append( clock_battery_text_area )
            display_group.append( clock_battery_group )
            print( "display status screen" )

            colors = [RED, GREEN, BLUE, OFF]
            for i in range (len(colors)):
                pixels.fill(colors[i])
                timenow = real_time_clock.datetime
                time_text_area.text = ("{:02d}:{:02d}:{:02d}Z".format( timenow.tm_hour, timenow.tm_min, timenow.tm_sec ))
                time.sleep( 0.75 )

            for i in range( 5 ):
                display_group.pop()

            del date_group
            del date_text
            del date_text_area

            del time_group
            del time_text
            del time_text_area

            del batch_group
            del batch_text
            del batch_text_area

            del main_battery_group
            del main_battery_text
            del main_battery_text_area

            del clock_battery_group
            del clock_battery_text
            del clock_battery_text_area
        else:
            # using the alt display
            name_group = displayio.Group( scale=3, x=10, y=25 )
            name_text = "STELLA"
            name_text_area = label.Label( terminalio.FONT, text=name_text, color=0xFFFFFF )
            name_group.append( name_text_area )
            display_group.append( name_group )

            subname_group = displayio.Group( scale=1, x=28, y=50 ) #x=45, y=60 )
            subname_text = "Air Quality"
            subname_text_area = label.Label( terminalio.FONT, text=subname_text, color=0xFFFFFF )
            subname_group.append( subname_text_area )
            display_group.append( subname_group )

            colors = [RED, GREEN, BLUE]
            for i in range (len(colors)):
                pixels.fill(colors[i])
                time.sleep( 0.75 )


            for i in range( 2 ):
                display_group.pop()

            UID_group = displayio.Group( scale=2, x=6, y=18 )
            UID_text = "UID:" + str(UID)
            UID_text_area = label.Label( terminalio.FONT, text=UID_text, color=0xFFFFFF )
            UID_group.append( UID_text_area )
            display_group.append( UID_group )

            version_group = displayio.Group( scale=2, x=6, y=45 )
            version_text = "Ver:" + SOFTWARE_VERSION_NUMBER
            version_text_area = label.Label( terminalio.FONT, text=version_text, color=0xFFFFFF )
            version_group.append( version_text_area )
            display_group.append( version_group )
            print( "display welcome screen" )

            for i in range (len(colors)):
                pixels.fill(colors[i])
                time.sleep( 0.75 )

            for i in range( 2 ):
                display_group.pop()

            del name_group
            del name_text
            del name_text_area

            del subname_group
            del subname_text
            del subname_text_area

            del UID_group
            del UID_text
            del UID_text_area

            del version_group
            del version_text
            del version_text_area

            gc.collect()
            timenow = real_time_clock.datetime

            date_group = displayio.Group( scale=2, x=5, y=20 )
            date_text = ("{}-{:02d}-{:02d}".format( timenow.tm_year, timenow.tm_mon, timenow.tm_mday ))
            date_text_area = label.Label( terminalio.FONT, text=date_text, color=0xFFFFFF )
            date_group.append( date_text_area )
            display_group.append( date_group )

            time_group = displayio.Group( scale=2, x=11, y=45 )
            time_text = ("{:02d}:{:02d}:{:02d}Z".format( timenow.tm_hour, timenow.tm_min, timenow.tm_sec ))
            time_text_area = label.Label( terminalio.FONT, text=time_text, color=0xFFFFFF )
            time_group.append( time_text_area )
            display_group.append( time_group )

            for i in range (len(colors)):
                pixels.fill(colors[i])
                timenow = real_time_clock.datetime
                time_text_area.text = ("{:02d}:{:02d}:{:02d}Z".format( timenow.tm_hour, timenow.tm_min, timenow.tm_sec ))
                time.sleep( 0.75 )

            for i in range( 2 ):
                display_group.pop()



            batch_group = displayio.Group( scale=2, x=15, y= 18)
            batch_text = "batch {}".format(batch_number)
            batch_text_area = label.Label( terminalio.FONT, text=batch_text, color=0xFFFFFF )
            batch_group.append( batch_text_area )
            display_group.append( batch_group )

            battery_voltage, battery_percentage = check_battery( vbat_voltage_pin, monitor )
            if battery_voltage < LOW_BATTERY_VOLTAGE:
                main_battery_status = "LOW"
            elif battery_voltage > 4.0:
                main_battery_status = "FULL"
            else:
                main_battery_status = " OK"
            if battery_percentage:
                main_battery_status = ( "{} %".format( int( battery_percentage )))


            main_battery_group = displayio.Group( scale=1, x=10, y=38 )
            main_battery_text = ("main battery:{}".format(main_battery_status))
            main_battery_text_area = label.Label( terminalio.FONT, text=main_battery_text, color=0xFFFFFF )
            main_battery_group.append( main_battery_text_area )
            display_group.append( main_battery_group )
            if clock_battery_low:
                clock_battery_status = "LOW"
            else:
                clock_battery_status = "OK"
            clock_battery_group = displayio.Group( scale=1, x=10, y=52 )
            clock_battery_text = ("clock battery: {}".format( clock_battery_status ))
            clock_battery_text_area = label.Label( terminalio.FONT, text=clock_battery_text, color=0xFFFFFF )
            clock_battery_group.append( clock_battery_text_area )
            display_group.append( clock_battery_group )
            print( "display status screen" )

            for i in range (len(colors)):
                pixels.fill(colors[i])
                time.sleep( 0.75 )


            for i in range( 3 ):
                display_group.pop()

            del date_group
            del date_text
            del date_text_area

            del time_group
            del time_text
            del time_text_area

            del batch_group
            del batch_text
            del batch_text_area

            del main_battery_group
            del main_battery_text
            del main_battery_text_area

            del clock_battery_group
            del clock_battery_text
            del clock_battery_text_area

        # create display structure
        gc.collect()
        if CO2_sensor:
            CO2_label_group = displayio.Group( scale=2, x=6, y=15)
            CO2_label_text = "CO"
            CO2_label_text_area = label.Label( terminalio.FONT, text=CO2_label_text, color=0xFFFFFF )
            CO2_label_group.append( CO2_label_text_area )
            display_group.append( CO2_label_group )

            CO2_subscript_group = displayio.Group( scale=1, x=30, y=20)
            CO2_subscript_text = "2"
            CO2_subscript_text_area = label.Label( terminalio.FONT, text=CO2_subscript_text, color=0xFFFFFF )
            CO2_subscript_group.append( CO2_subscript_text_area )
            display_group.append( CO2_subscript_group )

            CO2_value_group = displayio.Group( scale=2, x=42, y=15)
            CO2_value_text = "0ppm"
            CO2_value_text_area = label.Label( terminalio.FONT, text=CO2_value_text, color=0xFFFFFF )
            CO2_value_group.append( CO2_value_text_area )
            display_group.append( CO2_value_group )

        # temperature
        if WX_sensor and not alt_display:
            temperature_label_group = displayio.Group( scale=2, x=6, y=39)
            temperature_label_text = "Temp"
            temperature_label_text_area = label.Label( terminalio.FONT, text=temperature_label_text, color=0xFFFFFF )
            temperature_label_group.append( temperature_label_text_area )
            display_group.append( temperature_label_group )

            temperature_value_group = displayio.Group( scale=2, x=60, y=39)
            temperature_value_text = "0C"
            temperature_value_text_area = label.Label( terminalio.FONT, text= temperature_value_text, color=0xFFFFFF )
            temperature_value_group.append( temperature_value_text_area )
            display_group.append( temperature_value_group )

            # relative_humidity
            humidity_label_group = displayio.Group( scale=2, x=18, y=64)
            humidity_label_text = "RH"
            humidity_label_text_area = label.Label( terminalio.FONT, text=humidity_label_text, color=0xFFFFFF )
            humidity_label_group.append( humidity_label_text_area )
            display_group.append( humidity_label_group )

            humidity_value_group = displayio.Group( scale=2, x=60, y=64)
            humidity_value_text = "0%"
            humidity_value_text_area = label.Label( terminalio.FONT, text= humidity_value_text, color=0xFFFFFF )
            humidity_value_group.append( humidity_value_text_area )
            display_group.append( humidity_value_group )

            # barometric pressure

        # particulates line 1
        aqi_y = 87
        aqi_subscript_y = 92
        aqi_value_y = 110
        aqi_value_x = 10
        if alt_display:
            offset = 42
            aqi_y = aqi_y - offset
            aqi_subscript_y = aqi_subscript_y - offset
            aqi_value_y = aqi_y
            aqi_value_x = 44


        if pm25_sensor:
            aqi_label_group = displayio.Group( scale=2, x=10, y=aqi_y)
            aqi_label_text = "AQI"
            aqi_label_text_area = label.Label( terminalio.FONT, text=aqi_label_text, color=0xFFFFFF )
            aqi_label_group.append( aqi_label_text_area )
            display_group.append( aqi_label_group )

            aqi_subscript_group = displayio.Group( scale=1, x=45, y=aqi_subscript_y)
            aqi_subscript_text = "p"
            aqi_subscript_text_area = label.Label( terminalio.FONT, text=aqi_subscript_text, color=0xFFFFFF )
            aqi_subscript_group.append( aqi_subscript_text_area )
            display_group.append( aqi_subscript_group )

            aqi_value_group = displayio.Group( scale=2, x=aqi_value_x, y=aqi_value_y )
            aqi_value_text = "TBD"
            aqi_value_text_area = label.Label( terminalio.FONT, text=aqi_value_text, color=0xFFFFFF )
            aqi_value_group.append( aqi_value_text_area )
            display_group.append( aqi_value_group )

            #particles10_label_group = displayio.Group( scale=1, x=50, y=85)
            #particles10_label_text = "P1"
            #particles10_label_text_area = label.Label( terminalio.FONT, text=particles10_label_text, color=0xFFFFFF )
            #particles10_label_group.append( particles10_label_text_area )
            #display_group.append( particles10_label_group )

            #particles10_value_group = displayio.Group( scale=1, x=80, y=85)
            #particles10_value_text = "ug/m^3"
            #particles10_value_text_area = label.Label( terminalio.FONT, text= particles10_value_text, color=0xFFFFFF )
            #particles10_value_group.append( particles10_value_text_area )
            #display_group.append( particles10_value_group )

            particles_unit_group = displayio.Group( scale=1, x=86, y=85)
            particles_unit_text = "ug/m^3"
            particles_unit_text_area = label.Label( terminalio.FONT, text= particles_unit_text, color=0xFFFFFF )
            particles_unit_group.append( particles_unit_text_area )
            display_group.append( particles_unit_group )

            particles25_label_group = displayio.Group( scale=1, x=60, y=100)
            particles25_label_text = "P2.5"
            particles25_label_text_area = label.Label( terminalio.FONT, text=particles25_label_text, color=0xFFFFFF )
            particles25_label_group.append( particles25_label_text_area )
            display_group.append( particles25_label_group )

            particles25_value_group = displayio.Group( scale=1, x=90, y=100)
            particles25_value_text = "0"#ug/m^3"
            particles25_value_text_area = label.Label( terminalio.FONT, text= particles25_value_text, color=0xFFFFFF )
            particles25_value_group.append( particles25_value_text_area )
            display_group.append( particles25_value_group )

            # particulates line 2
            particles100_label_group = displayio.Group( scale=1, x=60, y=114)
            particles100_label_text = "P10"
            particles100_label_text_area = label.Label( terminalio.FONT, text=particles100_label_text, color=0xFFFFFF )
            particles100_label_group.append( particles100_label_text_area )
            display_group.append( particles100_label_group )

            particles100_value_group = displayio.Group( scale=1, x=90, y=114)
            particles100_value_text = "0"#ug/m^3"
            particles100_value_text_area = label.Label( terminalio.FONT, text= particles100_value_text, color=0xFFFFFF )
            particles100_value_group.append( particles100_value_text_area )
            display_group.append( particles100_value_group )

                # batch_number
        batch_border = 1
        batch_width = 29
        batch_height = 20
        batch_x = 54
        batch_y = 74
        if alt_display:
            batch_x = 93
            batch_y = 40
            batch_border = 2
            batch_width = 34
            batch_height =24
        border_color = 0xFFFFFF
        border = displayio.Palette(1)
        border[0] = border_color
        outer_rectangle = vectorio.Rectangle(pixel_shader=border, width=batch_width, height=batch_height, x=batch_x, y=batch_y)
        display_group.append( outer_rectangle )
        front_color = 0x000000
        front = displayio.Palette(1)
        front[0] = front_color
        inner_rectangle = vectorio.Rectangle(pixel_shader=front, width=batch_width-2*batch_border, height=batch_height-2*batch_border, x=batch_x+batch_border, y=batch_y+batch_border)
        display_group.append( inner_rectangle )

        batch_value_x = 57
        batch_value_y = 84
        if alt_display:
            batch_value_x = batch_x + 5
            batch_value_y = batch_y + 12

        batch_number_group = displayio.Group( scale=2, x=batch_value_x, y=batch_value_y)
        if int(batch_number) < 10:
            batch_number_text = (" " + batch_number )
        else:
            batch_number_text = batch_number
        batch_number_text_area = label.Label( terminalio.FONT, text=batch_number_text, color=0xFFFFFF )
        batch_number_group.append( batch_number_text_area )
        display_group.append( batch_number_group )

        del batch_border
        del batch_width
        del batch_height
        del batch_x
        del batch_y
        del border_color
        del border

    operational = True
    print()
    print("mirror data header:")
    print( header )
    # begin loop

    sample_interval_s = 1.0

    loop_count = 0
    max_loop_count = 0
    file_error_count = 0

    stale_CO2_data = True
    interval_start_time_h = 0
    reduced_datafile = create_AQ_reduced_data_file( desired_sample_interval_h, number_of_datapoints_to_average )
    reduced_data_sampling = False
    reduced_data_start_loop_count = 0

    try:
        while( operational ):
            gc.collect()
            loop_start_time_s = time.monotonic()
            gc.collect()
            mem_free_kB = gc.mem_free()/1000
            print( "free memory remaining = {} kB, {}%, loop {}, max {}, error count {}".format( mem_free_kB, (100* (mem_free_kB)/start_mem_free_kB ), loop_count, max_loop_count, file_error_count))
            # read clock
            if real_time_clock:
                timestamp = real_time_clock.datetime
            else:
                timestamp = null_time
            iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format(
                timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday,
                timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec )
            decimal_hour = timestamp_to_decimal_hour( timestamp )
            weekday = DAYS[ timestamp.tm_wday ]

            # read battery voltage
            battery_voltage, battery_percentage = check_battery(vbat_voltage_pin, monitor)
            #print("main battery voltage: {:.2f} V".format(battery_voltage))

            # read WX
            if WX_sensor:
                WX_reading = (WX_sensor.temperature, WX_sensor.relative_humidity, WX_sensor.pressure, WX_sensor.altitude)
            else:
                WX_reading = ( 0, 0, 0, 0 )
            #print( WX_reading )

            # read CO2
            if CO2_sensor:
                if CO2_sensor.data_ready:
                    stale_CO2_data = False
                    CO2_reading = ( CO2_sensor.CO2, CO2_sensor.temperature, CO2_sensor.relative_humidity, int( stale_CO2_data ))
                    stale_CO2_data = True
                    last_CO2_reading = ( CO2_sensor.CO2, CO2_sensor.temperature, CO2_sensor.relative_humidity, int( stale_CO2_data ))
                else:
                    CO2_reading = last_CO2_reading
            else:
                CO2_reading = ( 0, 0, 0, 0 )
            CO2_uncertainty = (50 + CO2_reading[0] * 0.05)
            #print( CO2_reading )

            # read PM
            if pm25_sensor:
                if not loop_count % 3: #measure particulates only every third loop
                    try:
                        stale_pm25_data = False
                        pm25_data = (pm25_sensor.read())
                        pm25_reading = (
                            pm25_data["pm10 standard"], pm25_data["pm25 standard"], pm25_data["pm100 standard"],
                            pm25_data["pm10 env"], pm25_data["pm25 env"], pm25_data["pm100 env"],
                            pm25_data["particles 03um"], pm25_data["particles 05um"], pm25_data["particles 10um"],
                            pm25_data["particles 25um"], pm25_data["particles 50um"], pm25_data["particles 100um"],
                            int(stale_pm25_data))
                        stale_pm25_data = True
                        last_pm25_reading = (
                            pm25_data["pm10 standard"], pm25_data["pm25 standard"], pm25_data["pm100 standard"],
                            pm25_data["pm10 env"], pm25_data["pm25 env"], pm25_data["pm100 env"],
                            pm25_data["particles 03um"], pm25_data["particles 05um"], pm25_data["particles 10um"],
                            pm25_data["particles 25um"], pm25_data["particles 50um"], pm25_data["particles 100um"],
                            int(stale_pm25_data))

                    except RuntimeError:
                        pm25_reading = last_pm25_reading
                #print( pm25_reading[1], pm25_reading[2] )
                aqi = calculate_aqi( pm25_reading[1], pm25_reading[2] )
                #print("AQI = {}".format(aqi))

            else:
                aqi = 0
                pm25_reading = (0,0,0,0,0,0,0,0,0,0,0,0,0)
            #print( pm25_reading )

            # display data
            gc.collect()
            if CO2_sensor:
                CO2_value_text_area.text = str(CO2_reading[0]) + "ppm"
            if not alt_display:
                if WX_sensor:
                    temperature_value_text_area.text = str(round(WX_reading[0],1)) + "C"
                    humidity_value_text_area.text = str(round(WX_reading[1],1)) + "%"
            if pm25_sensor:
                #particles10_value_text_area.text = str(round(pm25_reading[0],1)) + "ug/m^3"
                particles25_value_text_area.text = str(round(pm25_reading[1],1)) #+ "ug/m^3"
                particles100_value_text_area.text = str(round(pm25_reading[2],1)) # + "ug/m^3"
                if aqi < 100:
                    aqi_text = (" " + str(aqi))
                else:
                    aqi_text = str(aqi)
                aqi_value_text_area.text = aqi_text

            # build datapoint
            gc.collect()
            datapoint = ( DEVICE_TYPE )
            datapoint += str ( ", " )
            datapoint += ( SOFTWARE_VERSION_NUMBER )
            datapoint += str ( ", " )
            datapoint += str ( UID )
            datapoint += str ( ", " )
            datapoint += str ( batch_number )
            datapoint += str ( ", " )
            datapoint += str ( iso8601_utc )
            datapoint += str ( ", " )
            datapoint += str ( decimal_hour )
            datapoint += str ( ", " )
            datapoint += str ( CO2_reading[3] )
            datapoint += str ( ", " )
            datapoint += str ( CO2_reading[0] )
            datapoint += str ( ", " )
            datapoint += str ( int( CO2_uncertainty ))
            datapoint += str ( ", " )
            datapoint += str ( int( CO2_reading[1] ))
            datapoint += str ( ", " )
            datapoint += str ( CO2_sensor_air_temperature_uncertainty )
            datapoint += str ( ", " )
            datapoint += str ( int( CO2_reading[2] ))
            datapoint += str ( ", " )
            datapoint += str ( CO2_sensor_relative_humidity_uncertainty )
            datapoint += str ( ", " )
            datapoint += str ( int( WX_reading[0] ))
            datapoint += str ( ", " )
            datapoint += str ( WX_sensor_air_temperature_uncertainty )
            datapoint += str ( ", " )
            datapoint += str ( int( WX_reading[1] ))
            datapoint += str ( ", " )
            datapoint += str ( WX_sensor_relative_humidity_uncertainty )
            datapoint += str ( ", " )
            datapoint += str ( int( WX_reading[2] ))
            datapoint += str ( ", " )
            datapoint += str ( WX_sensor_barometric_pressure_uncertainty)
            datapoint += str ( ", " )
            datapoint += str ( WX_reading[3] )
            datapoint += str ( ", " )
            datapoint += str ( WX_sensor_altitude_uncertainty )
            datapoint += str ( ", " )
            datapoint += str ( aqi )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[0] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[1] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[2] )
            datapoint += str ( ", " )
            datapoint += str ( particle_concentration_uncertainty_ug )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[6] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[7] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[8] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[9] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[10] )
            datapoint += str ( ", " )
            datapoint += str ( pm25_reading[11] )
            datapoint += str ( ", " )
            datapoint += str ( "{:0.2f}".format( battery_voltage ))
            datapoint += str ( ", " )
            datapoint += str ( int(battery_percentage ))

            # record datapoint
            gc.collect()
            file_error_count = write_data_to_file( datapoint, pixels, file_error_count )
            # mirror data to usb
            gc.collect()
            print( datapoint )

            if decimal_hour - interval_start_time_h < 0:
                # check when decimal hour turns over at midnight, and reset the interval start time.
                interval_start_time_h = 0

            if decimal_hour - interval_start_time_h > desired_sample_interval_h:
                print( "time to take a reduced data sample" )
                interval_start_time_h = decimal_hour
                reduced_data_start_loop_count = loop_count
                reduced_data_sampling = True
                reduced_data_counter = 0
                data_accumulator = [ reduced_data_counter, 0, 0, 0, 0, 0 ]

            reduced_data_counter = loop_count - reduced_data_start_loop_count

            if reduced_data_sampling:
                print("reduced datapoint {}".format(reduced_data_counter))
                data_accumulator[0] = reduced_data_counter
                data_accumulator[1] = data_accumulator[1] + CO2_reading[0]
                data_accumulator[2] = data_accumulator[2] + aqi
                data_accumulator[3] = data_accumulator[3] + WX_reading[0]
                data_accumulator[4] = data_accumulator[4] + WX_reading[1]
                data_accumulator[5] = data_accumulator[5] + WX_reading[2]
                #print( data_accumulator )

            if reduced_data_counter > number_of_datapoints_to_average - 2:
                if reduced_data_sampling:
                    number_of_samples = data_accumulator[0] + 1
                    data_accumulator[1] = data_accumulator[1] / number_of_samples
                    data_accumulator[2] = data_accumulator[2] / number_of_samples
                    data_accumulator[3] = data_accumulator[3] / number_of_samples
                    data_accumulator[4] = data_accumulator[4] / number_of_samples
                    data_accumulator[5] = data_accumulator[5] / number_of_samples
                    #print( data_accumulator )
                    write_line_success = write_averages_to_reduced_data_file( reduced_datafile, UID, batch_number, iso8601_utc, decimal_hour, data_accumulator, pixels )
                    if write_line_success:
                        print( "averages written to file" )
                    else:
                        print( "failed to write averages to file" )
                reduced_data_sampling = False
            # wait to end of sample interval
            gc.collect()
            time_now_s = time.monotonic()
            if ( time_now_s - loop_start_time_s ) < sample_interval_s:
                interval_wait = True
            else:
                interval_wait = False

            while( interval_wait ):
                time_now_s = time.monotonic()
                if ( time_now_s - loop_start_time_s ) < sample_interval_s:
                    interval_wait = True
                else:
                    interval_wait = False
                time.sleep(0.05)
                #print( "waiting")
            loop_count += 1
    finally:
        sdcard.deinit()
        print( "sd card deinitialized" )
        displayio.release_displays()
        print( "displayio displays released" )
        i2c_bus.deinit()
        print( "i2c_bus deinitialized" )
        print("code exit: max loop count == {}, file error count == {}".format(max_loop_count, file_error_count))

def stall():
    print("intentionally stalled, press return to continue")
    input_string = False
    while input_string == False:
        input_string = input().strip()

def check_battery( pin, monitor ):
    battery_voltage = 0
    battery_percentage = 0
    if pin:
        #print( "getting analog voltage reading" )
        battery_voltage = get_voltage(pin)
    if monitor:
        #print( "getting fuel gauge battery health" )
        battery_voltage = monitor.cell_voltage          #wake up the fuel gauge chip with a couple of throw away commands
        battery_percentage = monitor.cell_percent
        time.sleep(0.1)
        battery_voltage = monitor.cell_voltage
        battery_percentage = monitor.cell_percent
    return battery_voltage, battery_percentage

def initialize_data_file(header):
    try:
        os.stat( DATA_FILE ) # fail if data.csv file is not already there
        #raise OSError # uncomment to force header everytime
        print( "data file already exists, does not need header" )
        return False
    except OSError:
        # setup the header for first time
        gc.collect()
        try:
            with open( DATA_FILE, "w" ) as f:
                f.write( header )
            print( "header written" )
        except OSError as err:
            print( err )
            pass
        return True

def write_data_to_file( datapoint, pixels, file_error_count):
    try:
        with open( DATA_FILE, "a" ) as f:
            if pixels:
                pixels.fill ( RED )
            f.write( str(datapoint ))
            f.write("\n")
            time.sleep( 0.05 )
            #f.close()
        if pixels:
            pixels.fill( OFF )
    except OSError as err:
        # TBD: maybe show something on the display like sd_full? this will "Error" every sample pass
        # "[Errno 30] Read-only filesystem" probably means no sd_card
        print( "\nError: sd card fail: {:}\n".format(err) )
        if pixels:
            pixels.fill( ORANGE ) #  ON to show error, likely no SD card present, or SD card full.
            time.sleep( 0.2 )
            pixels.fill( OFF )
        file_error_count += 1
    return file_error_count

def calculate_aqi( p25_reading, p100_reading ):
    gc.collect()
    p25_break_points = (0, 12, 12.1, 35.4, 35.5, 55.4, 55.5, 150.4, 150.5, 250.4, 250.5, 350.4, 350.5, 500.4)
    #print( p25_break_points )
    p100_break_points = (0, 54, 55, 154, 155, 254, 255, 354, 355, 424, 425, 504, 505, 604)
    #print( p100_break_points )
    AQI_levels = (0, 50, 51, 100, 101, 150, 151, 200, 201, 300, 301, 400, 401, 500)
    #print( AQI_levels )
    p25_index = 0
    p100_index = 0
    for i in range (len(AQI_levels)-1):
        if (p25_reading > p25_break_points[i] and p25_reading < p25_break_points[i+1]):
            p25_index = i
    #aqi = int((((ihi-ilo)/(bhi-blo))*(reading-blo))+ilo)
    aqi25 = int((((AQI_levels[p25_index+1] - AQI_levels[p25_index])/(p25_break_points[p25_index+1] - p25_break_points[p25_index]))*(p25_reading-p25_break_points[p25_index]))+AQI_levels[p25_index])
    #print(aqi25)
    for n in range (len(AQI_levels)-1):
        if (p100_reading > p100_break_points[i] and p100_reading < p100_break_points[i+1]):
            p100_index = i
    aqi100 = int((((AQI_levels[p100_index+1] - AQI_levels[p100_index])/(p100_break_points[p100_index+1] - p100_break_points[p100_index]))*(p100_reading-p25_break_points[p100_index]))+AQI_levels[p100_index])
    #print(aqi100)
    #print( p25_break_points[p25_index], p25_break_points[p25_index+1], AQI_levels[p25_index], AQI_levels[p25_index+1], p100_break_points[p100_index], p100_break_points[p100_index+1], AQI_levels[p100_index], AQI_levels[p100_index+1])
    if aqi100 > aqi25:
        return aqi100
    else:
        return aqi25

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def update_batch( timestamp ):
    gc.collect()
    datestamp = "{:04}{:02}{:02}".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday)
    try:
        with open( "/sd/batch.txt", "r" ) as b:
            try:
                previous_batchfile_string = b.readline()
                previous_datestamp = previous_batchfile_string[ 0:8 ]
                previous_batch_number = int( previous_batchfile_string[ 8: ])
            # TBD: catch error when /sd doesn't exist
            except ValueError:
                previous_batch_number = 0
                # corrupted data in batch number file, setting batch to 0
            if datestamp == previous_datestamp:
                # this is the same day, so increment the batch number
                batch_number = previous_batch_number + 1
            else:
                # this is a different day, so start the batch number at 0
                batch_number = 0
    except OSError:
            print( "batch.txt file not found" )
            batch_number = 0
    batch_string = ( "{:03}".format( batch_number ))
    batch_file_string = datestamp + batch_string
    try:
        with open( "/sd/batch.txt", "w" ) as b:
            b.write( batch_file_string )
        # TBD: catch error when /sd doesn't exist
    except OSError as err:
        print("Error: writing batch.txt {:}".format(err) )
        pass
    batch_string = ( "{:}".format( batch_number ))
    return batch_string

def get_voltage(pin):
    return (pin.value * 3.3) / 65536 * 2

def check_STELLA_config():
    desired_sample_interval_h, number_of_datapoints_to_average = 0.5, 10
    try:
        with open( "/sd/STELLA-AQ_reduce_config.txt", "r" ) as sc:
            try:
                header = sc.readline()
                #print( sampling_header )
                values = sc.readline()
                #print(sampling_values)
                values= values.rstrip('\n')
                values = values.split(',')
                #print(line)
                desired_sample_interval_h = float( values[ 0 ])
                number_of_datapoints_to_average = int( values[ 1 ])

            # TBD: catch error when /sd doesn't exist
            except ValueError:
                pass
    except OSError:
            print( "STELLA-AQ_reduce_config.txt file not found" )
            try:
                with open( "/sd/STELLA-AQ_reduce_config.txt", "w" ) as sc:
                    sc.write( "desired_sample_interval_h, number_of_datapoints_to_average\n" )
                    sc.write( "0.5, 10\n" )
                print("created new STELLA-AQ_reduce_config.txt file")
            except OSError:
                print("failed to create STELLA-AQ_reduce_config.txt file")
    return desired_sample_interval_h, number_of_datapoints_to_average

def create_AQ_reduced_data_file( desired_sample_interval_h, number_of_datapoints_to_average ):
    reduced_datafile = "/sd/AQ_reduced_data_{}h_{}p.csv".format( desired_sample_interval_h, number_of_datapoints_to_average )
    try:
        with open( reduced_datafile, "r" ) as f:
            try:
                header = f.readline()
                #print( sampling_header )
                values = f.readline()
                #print(sampling_values)
                values= values.rstrip('\n')
                values = values.split(',')
                #print(line)
                desired_sample_interval_h = float( values[ 0 ])
                number_of_datapoints_to_average = int( values[ 1 ])

            # TBD: catch error when /sd doesn't exist
            except ValueError:
                pass
    except OSError:
            print( "{} file not found".format( reduced_datafile ) )
            try:
                with open( reduced_datafile , "w" ) as f:
                    f.write( "UID, batch_number, timestamp_iso8601, dec_hour, CO2_ppm, AQIp, Air_temp, Air_humid, Air_press\n" )
                print("created new AQ_reduced_data_{}h_{}p.csv file".format( desired_sample_interval_h, number_of_datapoints_to_average ))
            except OSError:
                pass
    return reduced_datafile

def write_averages_to_reduced_data_file( reduced_datafile, UID, batch_number, timestamp_iso8601, decimal_hour, data_accumulator, pixels ):
    dataline = str(UID) + ", " +str(batch_number)+ ", " + timestamp_iso8601+ ", "
    dataline += str(decimal_hour)+ ", " + str(data_accumulator[1])+ ", " + str( data_accumulator[2])+ ", "
    dataline += str(data_accumulator[3])+ ", " + str(data_accumulator[4])+", " + str(data_accumulator[5])
    print(dataline)
    try:
        with open( reduced_datafile, "a" ) as avg:
            if pixels:
                pixels.fill ( BLUE )
            #UID, batch_number, timestamp, decimal_hour, CO2_ppm, AQIp, Air_temp, Air_humid, Air_press
            avg.write( dataline )
            avg.write("\n")
            time.sleep( 0.05 )
            #f.close()
            write_line_success = True
        if pixels:
            pixels.fill( OFF )
    except OSError as err:
        # TBD: maybe show something on the display like sd_full? this will "Error" every sample pass
        # "[Errno 30] Read-only filesystem" probably means no sd_card
        print( "\nError: sd card fail: {:}\n".format(err) )
        write_line_success = False
        if pixels:
            pixels.fill( ORANGE ) #  ON to show error, likely no SD card present, or SD card full.
            time.sleep( 0.2 )
            pixels.fill( OFF )
    return write_line_success

main()
